@@ -10,6 +10,8 @@ import openpaperwork_core.deps
 LOGGER = logging.getLogger(__name__)
 
 PDF_FILENAME = 'doc.pdf'
+PASSWD_FILENAME = 'passwd.txt'
+MAPPING_FILE = "page_map.csv"
 
 PDF_RENDER_FACTOR = 4
 
@@ -70,7 +72,6 @@ class PdfLineBox(object):
 
 
 class PdfPageMapping(object):
-    MAPPING_FILE = "page_map.csv"
     SEPARATOR = ","
 
     def __init__(self, plugin, doc_url):
@@ -79,7 +80,7 @@ class PdfPageMapping(object):
         self.doc_url = doc_url
 
         self.map_url = self.core.call_success(
-            "fs_join", self.doc_url, self.MAPPING_FILE
+            "fs_join", self.doc_url, MAPPING_FILE
         )
 
         self.mapping = None
@@ -104,15 +105,15 @@ class PdfPageMapping(object):
 
         if target_page_idx is not None:
             old_original = self.reverse_mapping.get(target_page_idx, None)
-            if old_original is not None:
+            if old_original is not None and old_original in self.mapping:
                 self.mapping.pop(old_original)
 
         old_target = self.mapping.get(original_page_idx, None)
-        if old_target is not None:
+        if old_target is not None and old_target in self.reverse_mapping:
             self.reverse_mapping.pop(old_target)
             self.page_mtimes[old_target] = now
 
-        if target_page_idx is None:
+        if target_page_idx is None and original_page_idx in self.mapping:
             self.mapping.pop(original_page_idx)
         else:
             self.mapping[original_page_idx] = target_page_idx
@@ -126,35 +127,46 @@ class PdfPageMapping(object):
         if len(self.mapping) <= 0:
             return
         for v in self.mapping.values():
+            if v is None:
+                continue
             if v >= self.nb_pages:
                 self.nb_pages = v + 1
 
     def load(self):
-        self.real_nb_pages = self.plugin._doc_internal_get_nb_pages_by_url(
-            self.doc_url, mapping=False
-        )
+        try:
+            self.real_nb_pages = self.plugin._doc_internal_get_nb_pages_by_url(
+                self.doc_url, mapping=False
+            )
 
-        self.mapping = {p: p for p in range(0, self.real_nb_pages)}
-        self.reverse_mapping = {p: p for p in range(0, self.real_nb_pages)}
-        now = datetime.datetime.now().timestamp()
-        self.page_mtimes = {p: now for p in range(0, self.real_nb_pages)}
+            self.mapping = {p: p for p in range(0, self.real_nb_pages)}
+            self.reverse_mapping = {p: p for p in range(0, self.real_nb_pages)}
+            now = datetime.datetime.now().timestamp()
+            self.page_mtimes = {p: now for p in range(0, self.real_nb_pages)}
+            if self.core.call_success("fs_exists", self.map_url) is None:
+                self.update_nb_pages()
+                return
+
+            with self.core.call_success("fs_open", self.map_url, "r") as fd:
+                # drop the first line
+                lines = fd.readlines()[1:]
+                lines = [line.split(",", 1) for line in lines]
+                lines = [
+                    (int(orig_page_idx), int(target_page_idx))
+                    for (orig_page_idx, target_page_idx) in lines
+                ]
+                lines.sort()
+                for (orig_page_idx, target_page_idx) in lines:
+                    if target_page_idx < 0:
+                        target_page_idx = None
+                    self.set_mapping(orig_page_idx, target_page_idx)
 
-        if self.core.call_success("fs_exists", self.map_url) is None:
             self.update_nb_pages()
-            return
-
-        with self.core.call_success("fs_open", self.map_url, "r") as fd:
-            # drop the first line
-            lines = fd.readlines()[1:]
-            for line in lines:
-                (orig_page_idx, target_page_idx) = line.split(",", 1)
-                orig_page_idx = int(orig_page_idx)
-                target_page_idx = int(target_page_idx)
-                if target_page_idx < 0:
-                    target_page_idx = None
-                self.set_mapping(orig_page_idx, target_page_idx)
-
-        self.update_nb_pages()
+        except Exception as exc:
+            LOGGER.critical(
+                "Failed to load page mapping for doc %s !",
+                self.doc_url, exc_info=exc
+            )
+            raise
 
     def load_reverse_only(self):
         """
@@ -220,9 +232,7 @@ class PdfPageMapping(object):
                 self.load()
         # Keep in mind we may have loaded only the mapping --> we don't know
         # how many pages there are in the PDF, so the mapping may be incomplete
-        original_page_idx = self.reverse_mapping.get(
-            target_page_idx, None
-        )
+        original_page_idx = self.reverse_mapping.get(target_page_idx, None)
         return original_page_idx
 
     def get_target_page_mtime(self, target_page_idx):
@@ -319,6 +329,8 @@ class PdfPageMapping(object):
 
 
 class Plugin(openpaperwork_core.PluginBase):
+    PRIORITY = 20
+
     def __init__(self):
         super().__init__()
         # we cache the hash of PDF files since they never change
@@ -328,10 +340,12 @@ class Plugin(openpaperwork_core.PluginBase):
         self.cache_nb_pages = {}
 
         self.cache_mappings = {}
+        self.cache_passwords = {}
 
     def get_interfaces(self):
         return [
             "doc_hash",
+            "doc_pdf",
             "doc_pdf_import",
             "doc_pdf_url",
             "doc_text",
@@ -361,7 +375,7 @@ class Plugin(openpaperwork_core.PluginBase):
             {
                 'interface': 'pillow',
                 'defaults': [
-                    'paperwork_backend.pillow.img',
+                    'openpaperwork_core.pillow.img',
                     'paperwork_backend.pillow.pdf',
                 ],
             },
@@ -369,6 +383,10 @@ class Plugin(openpaperwork_core.PluginBase):
                 'interface': 'poppler',
                 'defaults': ['paperwork_backend.poppler.memory'],
             },
+            {
+                'interface': 'urls',
+                'defaults': ['openpaperwork_core.urls'],
+            },
         ]
 
     def _get_pdf_url(self, doc_url):
@@ -379,6 +397,20 @@ class Plugin(openpaperwork_core.PluginBase):
             return None
         return pdf_url
 
+    def _get_pdf_password(self, doc_url):
+        password = self.cache_passwords.get(doc_url, None)
+        if password is not None:
+            return password
+
+        passwd_url = self.core.call_success(
+            "fs_join", doc_url, PASSWD_FILENAME
+        )
+        if self.core.call_success("fs_exists", passwd_url):
+            with self.core.call_success("fs_open", passwd_url, "r") as fd:
+                password = fd.read().strip()
+        self.cache_passwords[doc_url] = password
+        return password
+
     def _get_page_mapping(self, doc_url):
         if doc_url in self.cache_mappings:
             return self.cache_mappings[doc_url]
@@ -390,8 +422,12 @@ class Plugin(openpaperwork_core.PluginBase):
         pdf_url = self._get_pdf_url(doc_url)
         if pdf_url is None:
             return (None, None)
-        LOGGER.info("Opening %s", pdf_url)
-        doc = self.core.call_success("poppler_open", pdf_url)
+
+        password = self._get_pdf_password(doc_url)
+        LOGGER.info("Opening %s (password=%s)", pdf_url, bool(password))
+        doc = self.core.call_success(
+            "poppler_open", pdf_url, password=password
+        )
         return (pdf_url, doc)
 
     def is_doc(self, doc_url):
@@ -400,24 +436,33 @@ class Plugin(openpaperwork_core.PluginBase):
             return None
         return True
 
-    def doc_get_pdf_url_by_url(self, doc_url):
+    def doc_get_pdf_url_by_url(self, doc_url, write=False, allow_mapped=False):
+        if write:
+            return self.core.call_success("fs_join", doc_url, PDF_FILENAME)
+        if not allow_mapped:
+            mapping_url = self.core.call_success(
+                "fs_join", doc_url, MAPPING_FILE
+            )
+            if self.core.call_success("fs_exists", mapping_url):
+                return None
         return self._get_pdf_url(doc_url)
 
-    def doc_internal_get_hash_by_url(self, out: list, doc_url):
+    def flush_doc_cache(self, doc_url):
+        self.cache_hash.pop(doc_url, None)
+        self.cache_passwords.pop(doc_url, None)
+        self.cache_mappings.pop(doc_url, None)
+        self.cache_nb_pages.pop(doc_url, None)
+
+    def doc_get_hash_by_url(self, doc_url):
         pdf_url = self._get_pdf_url(doc_url)
         if pdf_url is None:
             return
-        mapping = self._get_page_mapping(doc_url)
-        h = mapping.get_map_hash()
-        if h is not None:
-            out.append(h)
 
         # cache the hash of doc.pdf to speed up imports
-        cache_key = "hash_{}".format(doc_url)
-        if cache_key not in self.cache_hash:
+        if doc_url not in self.cache_hash:
             h = self.core.call_success("fs_hash", pdf_url)
-            self.cache_hash[cache_key] = h
-        out.append(self.cache_hash[cache_key])
+            self.cache_hash[doc_url] = h
+        return self.cache_hash[doc_url]
 
     def page_internal_get_hash_by_url(self, out: list, doc_url, page_idx):
         pdf_url = self._get_pdf_url(doc_url)
@@ -431,11 +476,10 @@ class Plugin(openpaperwork_core.PluginBase):
         out.append(page_hash)
 
         # cache the hash of doc.pdf to speed up imports
-        cache_key = "hash_{}".format(doc_url)
-        if cache_key not in self.cache_hash:
+        if doc_url not in self.cache_hash:
             h = self.core.call_success("fs_hash", pdf_url)
-            self.cache_hash[cache_key] = h
-        out.append(self.cache_hash[cache_key])
+            self.cache_hash[doc_url] = h
+        out.append(self.cache_hash[doc_url])
 
     def doc_internal_get_mtime_by_url(self, out: list, doc_url):
         pdf_url = self._get_pdf_url(doc_url)
@@ -473,6 +517,12 @@ class Plugin(openpaperwork_core.PluginBase):
         nb_pages = pdf.get_n_pages()
         return nb_pages
 
+    def doc_pdf_get_real_nb_pages_by_url(self, doc_url):
+        # Poppler is not thread-safe
+        return self.core.call_one(
+            "mainloop_execute", self._doc_get_real_nb_pages_by_url, doc_url
+        )
+
     def _doc_internal_get_nb_pages_by_url(
             self, doc_url, mapping=True):
         if mapping:
@@ -487,10 +537,7 @@ class Plugin(openpaperwork_core.PluginBase):
         if doc_url in self.cache_nb_pages:
             r = self.cache_nb_pages[doc_url]
         else:
-            # Poppler is not thread-safe
-            r = self.core.call_one(
-                "mainloop_execute", self._doc_get_real_nb_pages_by_url, doc_url
-            )
+            r = self.doc_pdf_get_real_nb_pages_by_url(doc_url)
             if r is not None:
                 self.cache_nb_pages[doc_url] = r
         return r if r is not None else 0
@@ -501,7 +548,7 @@ class Plugin(openpaperwork_core.PluginBase):
             return
         out.append(r)
 
-    def page_get_img_url(self, doc_url, page_idx, write=False):
+    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
         if write:
             return None
         pdf_url = self._get_pdf_url(doc_url)
@@ -520,17 +567,35 @@ class Plugin(openpaperwork_core.PluginBase):
             page_idx, reverse_only=(page_idx == 0)
         )
         if original_page_idx is None:
-            if page_idx != 0:
+            if page_idx != 0 and page_idx >= mapping.nb_pages:
                 return None
-            else:
-                original_page_idx = page_idx
+            original_page_idx = page_idx
+
+        password = self._get_pdf_password(doc_url)
+        if password is not None:
+            password = password.encode("utf-8").hex()
+
         # same URL used in browsers
-        return "{}#page={}".format(pdf_url, str(original_page_idx + 1))
+        url = self.core.call_success(
+            "url_args_join", pdf_url,
+            page=str(original_page_idx + 1),
+            password=password
+        )
+        return url
 
     @staticmethod
-    def _custom_split(input_str, input_rects, splitter):
+    def _custom_split(input_str, input_rects, splitter, log_txt):
         # turn text and layout from Poppler into boxes
-        assert(len(input_str) == len(input_rects))
+        # XXX(Jflesch): following assert fails sometimes ? oO
+        # assert len(input_str) == len(input_rects)
+        if len(input_str) != len(input_rects):
+            LOGGER.warning(
+                "%s: Input strings: %d ; Input rects: %d",
+                log_txt, len(input_str), len(input_rects)
+            )
+            m = min(len(input_str), len(input_rects))
+            input_str = input_str[:m]
+            input_rects = input_rects[:m]
         input_el = zip(input_str, input_rects)
         for (is_split, group) in itertools.groupby(
                     input_el,
@@ -543,7 +608,7 @@ class Plugin(openpaperwork_core.PluginBase):
             for (letter, rect) in group:
                 letters += letter
                 rects.append(rect)
-            yield(letters, rects)
+            yield (letters, rects)
 
     def _doc_get_text_by_url(self, out: list, doc_url):
         task = "pdf_get_text_by_url({})".format(doc_url)
@@ -559,8 +624,18 @@ class Plugin(openpaperwork_core.PluginBase):
             if not mapping.has_original_page_idx(page_idx):
                 continue
             page = pdf.get_page(page_idx)
-            txt = page.get_text()
-            txt = txt.strip()
+            # some PDF are really badly damaged
+            if page is None:
+                continue
+            try:
+                txt = page.get_text()
+                txt = txt.strip()
+            except UnicodeDecodeError as exc:
+                LOGGER.warning(
+                    "%s p%d: UnicodeDecodeError: Assuming page has no text",
+                    doc_url, page_idx, exc_info=exc
+                )
+                continue
             if txt == "":
                 continue
             out.append(txt)
@@ -589,7 +664,14 @@ class Plugin(openpaperwork_core.PluginBase):
         if page is None:
             return None
 
-        return len(page.get_text().strip()) > 0
+        try:
+            return len(page.get_text().strip()) > 0
+        except UnicodeDecodeError as exc:
+            LOGGER.warning(
+                "%s p%d: UnicodeDecodeError: Assuming page has no text",
+                doc_url, page_idx, exc_info=exc
+            )
+            return False
 
     def page_has_text_by_url(self, doc_url, page_idx):
         pdf_url = self._get_pdf_url(doc_url)
@@ -604,6 +686,41 @@ class Plugin(openpaperwork_core.PluginBase):
             "mainloop_execute", self._page_has_text_by_url, doc_url, page_idx
         )
 
+    def _page_get_text_by_url(self, doc_url, page_idx):
+        if doc_url in self.cache_nb_pages:
+            if page_idx >= self.cache_nb_pages[doc_url]:
+                return None
+
+        (pdf_url, pdf) = self._open_pdf(doc_url)
+        if pdf is None:
+            return None
+
+        page = pdf.get_page(page_idx)
+        if page is None:
+            return None
+
+        try:
+            return page.get_text().strip()
+        except UnicodeDecodeError as exc:
+            LOGGER.warning(
+                "%s p%d: UnicodeDecodeError: Assuming page has no text",
+                doc_url, page_idx, exc_info=exc
+            )
+            return False
+
+    def page_get_text_by_url(self, doc_url, page_idx):
+        pdf_url = self._get_pdf_url(doc_url)
+        if pdf_url is None:
+            return None
+        mapping = self._get_page_mapping(doc_url)
+        page_idx = mapping.get_original_page_idx(page_idx)
+        if page_idx is None:
+            return None
+        # Poppler is not thread-safe
+        return self.core.call_one(
+            "mainloop_execute", self._page_get_text_by_url, doc_url, page_idx
+        )
+
     def _page_get_boxes_by_url(self, doc_url, page_idx):
         if doc_url in self.cache_nb_pages:
             if page_idx >= self.cache_nb_pages[doc_url]:
@@ -617,8 +734,15 @@ class Plugin(openpaperwork_core.PluginBase):
         if pdf_page is None:
             return
 
-        txt = pdf_page.get_text()
-        if txt.strip() == "":
+        try:
+            txt = pdf_page.get_text()
+            if txt.strip() == "":
+                return None
+        except UnicodeDecodeError as exc:
+            LOGGER.warning(
+                "%s p%d: UnicodeDecodeError: Assuming page has no text",
+                doc_url, page_idx, exc_info=exc
+            )
             return None
 
         layout = pdf_page.get_text_layout()
@@ -628,11 +752,11 @@ class Plugin(openpaperwork_core.PluginBase):
 
         line_boxes = []
         for (line, line_rects) in self._custom_split(
-                    txt, layout, lambda x: x == "\n"
+                    txt, layout, lambda x: x == "\n", "lines",
                 ):
             words = []
             for (word, word_rects) in self._custom_split(
-                        line, line_rects, lambda x: x.isspace()
+                        line, line_rects, lambda x: x.isspace(), "words"
                     ):
                 word_box = PdfWordBox(word, word_rects)
                 words.append(word_box)
@@ -652,8 +776,22 @@ class Plugin(openpaperwork_core.PluginBase):
             "mainloop_execute", self._page_get_boxes_by_url, doc_url, page_idx
         )
 
-    def doc_pdf_import(self, src_file_uri):
-        (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
+    def doc_pdf_import(self, src_file_uri, password=None, target_doc_id=None):
+        doc_id = None
+        doc_url = None
+
+        if target_doc_id is not None:
+            doc_id = target_doc_id
+            doc_url = self.core.call_success(
+                "doc_id_to_url", doc_id, existing=False
+            )
+            if doc_url is None:
+                doc_id = None
+            elif self.core.call_success("fs_exists", doc_url):
+                doc_id = None
+
+        if doc_id is None:
+            (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
 
         # just to be safe
         self.cache_mappings.pop(doc_url, None)
@@ -663,9 +801,18 @@ class Plugin(openpaperwork_core.PluginBase):
         self.core.call_success("fs_mkdir_p", doc_url)
         self.core.call_success("fs_copy", src_file_uri, pdf_url)
 
+        if password is not None:
+            passwd_url = self.core.call_success(
+                "fs_join", doc_url, PASSWD_FILENAME
+            )
+            with self.core.call_success("fs_open", passwd_url, "w") as fd:
+                fd.write(password)
+
         try:
             # check the PDF is readable
-            doc = self.core.call_success("poppler_open", pdf_url)
+            doc = self.core.call_success(
+                "poppler_open", pdf_url, password=password
+            )
             exc_info = None
         except Exception as exc:
             doc = None
@@ -677,6 +824,7 @@ class Plugin(openpaperwork_core.PluginBase):
             else:
                 LOGGER.error("Failed to read %s", pdf_url, exc_info=exc_info)
             self.core.call_success("fs_rm_rf", doc_url, trash=False)
+            return (None, None)
 
         return (doc_id, doc_url)
 
@@ -757,12 +905,11 @@ class Plugin(openpaperwork_core.PluginBase):
             original_source_page_idx = source_mapping.get_original_page_idx(
                 target_source_page_idx
             )
-            if original_source_page_idx is None:
-                # this page is not handled by us ; still, we must shift
-                # all our pages
-                original_source_page_idx = target_source_page_idx
+            # if original_source_page_idx is None means
+            # this page is not handled by us ; still, we must shift
+            # all our pages
             LOGGER.info(
-                "- Removing page (original=%d, target=%d) from %s",
+                "- Removing page (original=%s, target=%d) from %s",
                 original_source_page_idx, target_source_page_idx,
                 source_doc_url
             )
@@ -776,39 +923,45 @@ class Plugin(openpaperwork_core.PluginBase):
             dest_mapping.make_room_for_target_page(target_dest_page_idx)
 
         if source_doc_url == dest_doc_url:
-            assert(source_mapping == dest_mapping)
-            LOGGER.info(
-                "New mapping: %s: original=p%d --> target=p%d",
-                source_doc_url, original_source_page_idx, target_dest_page_idx
-            )
-            source_mapping.set_mapping(
-                original_source_page_idx, target_dest_page_idx
-            )
-            source_mapping.update_nb_pages()
+            assert source_mapping is dest_mapping
+            if original_source_page_idx is not None:
+                LOGGER.info(
+                    "New mapping: %s: original=p%d --> target=p%d",
+                    source_doc_url, original_source_page_idx,
+                    target_dest_page_idx
+                )
+                source_mapping.set_mapping(
+                    original_source_page_idx, target_dest_page_idx
+                )
+
+                source_mapping.update_nb_pages()
         elif source_is_pdf:
-            # export the PDF page as an image file
-            # it relies on other model plugins (interface 'page_img'), but they
-            # can't be declared as dependencies, as we do provide 'page_img'
-            # too. It would make a dependency loop.
+            if original_source_page_idx is not None:
+                # export the PDF page as an image file
+                # it relies on other model plugins (interface 'page_img'), but
+                # they can't be declared as dependencies, as we do provide
+                # 'page_img' too. It would make a dependency loop.
 
-            # we are a low priority plugin: other plugins should
-            # already have made room for our page if required
+                # we are a low priority plugin: other plugins should
+                # already have made room for our page if required
 
-            source_img = "{}#page={}".format(
-                source_pdf_url, str(original_source_page_idx + 1)
-            )
-            LOGGER.info("Generating image from %s", source_img)
-            source_img = self.core.call_success("url_to_pillow", source_img)
+                source_img = "{}#page={}".format(
+                    source_pdf_url, str(original_source_page_idx + 1)
+                )
+                LOGGER.info("Generating image from %s", source_img)
+                source_img = self.core.call_success(
+                    "url_to_pillow", source_img
+                )
 
-            # since PDF are not writable by themselves, we can call
-            # page_get_img_url(write=True). We are sure that our
-            # implementation of this method won't reply
-            dest_img = self.core.call_success(
-                "page_get_img_url", dest_doc_url, target_dest_page_idx,
-                write=True
-            )
-            LOGGER.info("Writting page image back as %s", dest_img)
-            self.core.call_success("pillow_to_url", source_img, dest_img)
+                # since PDF are not writable by themselves, we can call
+                # page_get_img_url(write=True). We are sure that our
+                # implementation of this method won't reply
+                dest_img = self.core.call_success(
+                    "page_get_img_url", dest_doc_url, target_dest_page_idx,
+                    write=True, original=True
+                )
+                LOGGER.info("Writting page image back as %s", dest_img)
+                self.core.call_success("pillow_to_url", source_img, dest_img)
 
         if source_is_pdf:
             source_mapping.save()
